/**
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 2010 France Telecom R&D
 * Contact: jonas-team@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 *
 * --------------------------------------------------------------------------
 * $Id: AbsReliableJMSTransportMessageDrivenBean.java 21483 2011-06-23 14:47:13Z jppouzol $
 * --------------------------------------------------------------------------
 */

package org.ow2.jonas.cxf.transport.reliablejms;

import java.io.UnsupportedEncodingException;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.TreeSet;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.jms.JMSException;
import javax.jms.Message;
import javax.jms.MessageListener;
import javax.jms.Queue;
import javax.xml.namespace.QName;

import org.apache.cxf.Bus;
import org.apache.cxf.BusFactory;
import org.apache.cxf.binding.soap.SoapFault;
import org.apache.cxf.binding.soap.SoapMessage;
import org.apache.cxf.common.logging.LogUtils;
import org.apache.cxf.helpers.DOMUtils;
import org.apache.cxf.interceptor.ClientFaultConverter;
import org.apache.cxf.interceptor.InterceptorChain;
import org.apache.cxf.jaxws.interceptors.HolderInInterceptor;
import org.apache.cxf.message.Exchange;
import org.apache.cxf.message.ExchangeImpl;
import org.apache.cxf.message.FaultMode;
import org.apache.cxf.message.MessageImpl;
import org.apache.cxf.phase.Phase;
import org.apache.cxf.phase.PhaseInterceptorChain;
import org.apache.cxf.service.model.BindingFaultInfo;
import org.apache.cxf.service.model.BindingInfo;
import org.apache.cxf.service.model.BindingOperationInfo;
import org.apache.cxf.transport.Destination;
import org.apache.cxf.transport.jms.JMSConfiguration;
import org.apache.cxf.transport.jms.JMSConstants;
import org.apache.cxf.transport.jms.JMSMessageHeadersType;
import org.apache.cxf.transport.jms.JMSUtils;
import org.ow2.jonas.cxf.transport.reliablejms.api.AsyncHandler;
import org.ow2.jonas.cxf.transport.reliablejms.api.ContextInfo;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

public abstract class AbsReliableJMSTransportMessageDrivenBean implements MessageListener {

    /**
     * Logger.
     */
    private static final Logger LOG = LogUtils.getL7dLogger(AbsReliableJMSTransportMessageDrivenBean.class);

    private IReliableJMSTransportSingletonBean singletonBean = null;

    protected void setSingletonBean(IReliableJMSTransportSingletonBean singletonBean) {
        this.singletonBean = singletonBean;
    }

    protected Destination getDestination(String queueName) {
        if (singletonBean == null) {
            throw new IllegalStateException("Singleton bean not injected");
        }
        return singletonBean.getDestination(queueName);
    }

    protected Bus getBus() {
        if (singletonBean == null) {
            throw new IllegalStateException("Singleton bean not injected");
        }
        return singletonBean.getBus();
    }

    
    
    protected void processInMessage(org.apache.cxf.message.Message inMessage) {

        // Get destination/configuration/mode from the message
        Destination destination = inMessage.getDestination();
        ReliableJMSTransportDestination reliableJMSTransportDestination = null;
        ReliableJMSTransportConfiguration reliableJMSTransportConfiguration = null;
        ReliableJMSTransportModeType mode = null;

        if (destination != null && destination instanceof ReliableJMSTransportDestination) {
            reliableJMSTransportDestination = (ReliableJMSTransportDestination) destination;
            reliableJMSTransportConfiguration = reliableJMSTransportDestination.getConfiguration();
            mode = reliableJMSTransportDestination.getMode();
        } else {
            throw new IllegalStateException("Invalid destination. Found '" + destination + "'.");
        }

        // switch mode
        switch (mode) {
        case CLIENT_RECEIVER:

            // Inbound message
            inMessage.put(org.apache.cxf.message.Message.INBOUND_MESSAGE, true);

            // True because we're handling a response
            inMessage.put(org.apache.cxf.message.Message.REQUESTOR_ROLE, true);

            // Build dummy output message that needs to be a SOAP message
            SoapMessage outMessage = new SoapMessage(new MessageImpl());

            // Define empty holders (to avoid any NPE later in the interceptors)
            outMessage.put(HolderInInterceptor.CLIENT_HOLDERS, Collections.<Object> emptyList());

            // Create exchange objet and make links between input and output
            // messages
            Exchange exchange = new ExchangeImpl();
            exchange.setInMessage(inMessage);
            exchange.setOutMessage(outMessage);
            inMessage.setExchange(exchange);
            outMessage.setExchange(exchange);


            // disable output interceptors so that there are no calls on the
            // output message as we don't want one
            // NOTE: oneWay is not working as types are in the outputMessage of
            // the operationInfo. Resetting them will make the unwrapped fail.
            InterceptorChain interceptorChain = new PhaseInterceptorChain(new TreeSet<Phase>());
            outMessage.setInterceptorChain(interceptorChain);

            // Store destination configuration in the exchange object
            exchange.put(ReliableJMSTransportConfiguration.class, reliableJMSTransportConfiguration);
            break;
        default:
        }
        inMessage.put(ReliableJMSTransportModeType.class, mode);
    }

    public void onMessage(Message message) {

        // Get name of the queue on which this MDB is listening to
        javax.jms.Destination jmsDestination;
        try {
            jmsDestination = message.getJMSDestination();
        } catch (JMSException e) {
            throw new IllegalStateException("Unable to get the JMS destination of the given message '" + message + "'.", e);
        }
        Queue jmsQueueDestination = null;
        if (jmsDestination instanceof Queue) {
            jmsQueueDestination = (Queue) jmsDestination;
        } else {
            throw new IllegalStateException("The message destination for the message '" + message + "' is not a queue.");
        }
        String queueName;
        try {
            queueName = jmsQueueDestination.getQueueName();
        } catch (JMSException e) {
            throw new IllegalStateException(
                    "Unable to get the JMS queue name  of the given queue '" + jmsQueueDestination + "'.", e);
        }

        javax.jms.Destination jmsReplyTo = null;
        try {
            jmsReplyTo = message.getJMSReplyTo();
        } catch (JMSException e) {
            throw new IllegalStateException(
                    "Unable to get the JMS replyto of the given queue '" + jmsQueueDestination + "'.", e);
        }
        

        // Get the destination object that was configured for this given queue
        // name
        Destination destination = getDestination(queueName);
        if (destination == null) {
            throw new IllegalStateException("Cannot get the destination object configured for the given queue '" + queueName
                    + "'.");
        }

        // Get Correlation ID
        String correlationID = null;
        try {
            correlationID = message.getJMSCorrelationID();
        } catch (JMSException e) {
            throw new IllegalStateException("Cannot get correlation ID", e);
        }

        LOG.fine("Listening on the destination '" + queueName + "' using address '" + destination.getAddress() + "'.");

        // create an empty configuration
        JMSConfiguration jmsConfig = new JMSConfiguration();

        // Create a CXF message that will be populated by the JMS message
        MessageImpl inMessage = new MessageImpl();
        try {
            JMSUtils.populateIncomingContext(message, inMessage, JMSConstants.JMS_SERVER_REQUEST_HEADERS, jmsConfig);
        } catch (UnsupportedEncodingException e) {
            throw new IllegalStateException("Cannot create CXF message using the given JMS message", e);
        }

        // Payload
        try {
            JMSUtils.retrieveAndSetPayload(inMessage, message, (String) inMessage.get(org.apache.cxf.message.Message.ENCODING));
        } catch (UnsupportedEncodingException e) {
            throw new IllegalStateException("Cannot get payload of the given JMS message", e);
        }

        // save replyTo
        if (jmsReplyTo != null) {
            inMessage.put("JMSReplyTo", jmsReplyTo);
        }

        inMessage.put("CorrelationID", correlationID);
        inMessage.put(JMSConstants.JMS_SERVER_RESPONSE_HEADERS, new JMSMessageHeadersType());
        inMessage.put(JMSConstants.JMS_REQUEST_MESSAGE, message);
        inMessage.setDestination(destination);

        // Add properties in the message (if any)
        processInMessage(inMessage);
        


        try {
            BusFactory.setThreadDefaultBus(getBus());

            // Call the invoker
            destination.getMessageObserver().onMessage(inMessage);
        } finally {
            BusFactory.setThreadDefaultBus(null);
            Exchange exchange = inMessage.getExchange();

            if (ReliableJMSTransportModeType.CLIENT_RECEIVER == inMessage.get(ReliableJMSTransportModeType.class)) {

                // Check if object received is an exception and then send it to async handlers
                org.apache.cxf.message.Message receivedExceptionMessage = exchange.getInFaultMessage();
                if (receivedExceptionMessage != null) {
                    ReliableJMSTransportConfiguration reliableJMSTransportConfiguration = exchange.get(ReliableJMSTransportConfiguration.class);
                    if (reliableJMSTransportConfiguration == null) {
                        throw new IllegalStateException("Unable to get the MDB receiver configuration for operation '"
                                + inMessage.getExchange() + "'.");
                    }

                    // For the given fault, find a operation using this fault and store the BindingOperationInfo on the exchange
                    // object
                    // This value will be analyzed by the ClientFaultConverter class in order to do the mapping with the real
                    // exception class and the fault details
                    Exception exception = receivedExceptionMessage.getContent(Exception.class);
                    SoapFault fault = null;
                    BindingOperationInfo matchingBindingOperationInfo = null;
                    if (exception instanceof SoapFault) {
                        fault = (SoapFault) exception;
                    }
                    if (fault != null && fault.getDetail() != null) {
                        // Unmarshal exception
                        Element exDetail = (Element) DOMUtils.getChild(fault.getDetail(), Node.ELEMENT_NODE);
                        
                        // Try to see if the JMS Message has the operation name
                        String operationName = null;
                        try {
                            operationName = message.getStringProperty(org.apache.cxf.message.Message.WSDL_OPERATION);
                        } catch (JMSException e) {
                            LOG.log(Level.SEVERE, "Unable to get the operation name from the JMS Message", e);
                        }
                        
                        
                        // now, search this Exception QName in one of the operations of the WSDL
                        if (exDetail != null) {
                            QName qname = new QName(exDetail.getNamespaceURI(), exDetail.getLocalName());
                            BindingInfo bindingInfo = ((ReliableJMSTransportDestination) destination).getEndpointInfo().getBinding();
                            for (BindingOperationInfo b : bindingInfo.getOperations()) {
                                // look at the faults and stop when there is a match
                                Collection<BindingFaultInfo> bindingFaults = b.getFaults();
                                if (bindingFaults != null) {
                                    for (BindingFaultInfo bindingFaultInfo : bindingFaults) {
                                        if (bindingFaultInfo.getFaultInfo().getMessageParts().get(0).getConcreteName().equals(
                                                qname)) {
                                            // Check the operation name
                                            if (operationName != null) {
                                                String currentOperationName = b.getOperationInfo().getName().toString();
                                                if (operationName.equals(currentOperationName)) {
                                                    // Got a matching operation for this fault !
                                                    matchingBindingOperationInfo = b;
                                                    break;
                                                }
                                            } else {
                                                // Got a matching operation for this fault !
                                                matchingBindingOperationInfo = b;
                                                break;
                                            }
                                        }

                                    }
                                }
                                // Stop if found
                                if (matchingBindingOperationInfo != null) {
                                    break;
                                }
                            }
                            // Store it if found
                            if (matchingBindingOperationInfo != null) {
                                exchange.put(BindingOperationInfo.class, matchingBindingOperationInfo);
                            }

                            // Convert previous unmapped fault to a real Exception object
                            new ClientFaultConverter().handleMessage(receivedExceptionMessage);

                        }
                    }
                    // Read again the fault (may have been updated)
                    exception = receivedExceptionMessage.getContent(Exception.class);

                    // Now, notify the clients handlers that there is a fault
                    List<AsyncHandler> asyncHandlers = reliableJMSTransportConfiguration.getAsyncHandlers();
                    if (asyncHandlers != null) {
                        boolean distributed = false;
                        for (AsyncHandler asyncHandler : asyncHandlers) {
                            ContextInfo contextInfo = new ContextInfo();
                            contextInfo.setCorrelationID(correlationID);
                            if (matchingBindingOperationInfo != null) {
                                contextInfo.setWsdlOperation(matchingBindingOperationInfo.getName());
                                contextInfo.setWsdlService((QName) exchange.get("javax.xml.ws.wsdl.service"));
                                contextInfo.setWsdlInterface((QName) exchange.get("javax.xml.ws.wsdl.interface"));
                                contextInfo.setWsdlPort((QName) exchange.get("javax.xml.ws.wsdl.port"));
                            }

                            boolean known = false;
                            try {
                                known = asyncHandler.doYouKnowThisContext(contextInfo);
                            } catch (Error e) {
                                LOG.log(Level.SEVERE, "Error while asking the asynchandler. continue", e);
                            } catch (Exception e) {
                                LOG.log(Level.SEVERE, "Error while asking the asynchandler. continue", e);
                            }
                            
                            if (known) {
                                asyncHandler.doHandlerFault(exception, contextInfo);
                                distributed = true;
                            }

                        }
                        if (!distributed) {
                            throw new IllegalStateException("Message has not been processed as no defined handlers was accepting it. Trying to send the message again...And then send it to the DMQ (if configured)");
                        }                    
                    }
                }
            }
            if (ReliableJMSTransportModeType.SERVER_ENDPOINT == inMessage.get(ReliableJMSTransportModeType.class)) {
                // Check if there was a failure in order to rethrow this exception
                Exception exception = inMessage.getContent(Exception.class);
                if (exception != null) {
                    FaultMode faultMode = (FaultMode) inMessage.getExchange().getOutFaultMessage().get(org.apache.cxf.message.FaultMode.class);
                    if (faultMode == FaultMode.CHECKED_APPLICATION_FAULT) {
                        LOG.fine("There is a checked exception that has been sent to the client");
                    } else {
                        LOG.fine("There is an unchecked exception. Not sending back to the client, the JMS message will be redelivered...");

                        // Check cause
                        Throwable cause = exception.getCause();
                        if (cause instanceof RuntimeException) {
                            throw (RuntimeException) cause;
                        } else {
                            throw new IllegalStateException("Error while processing the message", cause);
                        }
                    }
                }
            }
        }

    }

}
