/**
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 2010 France Telecom R&D
 * Contact: jonas-team@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 *
 * --------------------------------------------------------------------------
 * $Id: ReliableJMSTransportConduit.java 21483 2011-06-23 14:47:13Z jppouzol $
 * --------------------------------------------------------------------------
 */

package org.ow2.jonas.cxf.transport.reliablejms;

import java.io.IOException;
import java.io.OutputStream;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.UUID;
import java.util.concurrent.atomic.AtomicLong;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.jms.Connection;
import javax.jms.Destination;
import javax.jms.JMSException;
import javax.jms.MessageProducer;
import javax.jms.Session;
import javax.xml.namespace.QName;

import org.apache.cxf.common.logging.LogUtils;
import org.apache.cxf.configuration.ConfigurationException;
import org.apache.cxf.endpoint.Client;
import org.apache.cxf.helpers.CastUtils;
import org.apache.cxf.io.CachedOutputStream;
import org.apache.cxf.io.CachedOutputStreamCallback;
import org.apache.cxf.message.Exchange;
import org.apache.cxf.message.FaultMode;
import org.apache.cxf.message.Message;
import org.apache.cxf.message.MessageImpl;
import org.apache.cxf.message.MessageUtils;
import org.apache.cxf.transport.AbstractConduit;
import org.apache.cxf.transport.Conduit;
import org.apache.cxf.transport.jms.JMSConfiguration;
import org.apache.cxf.transport.jms.JMSUtils;
import org.apache.cxf.ws.addressing.EndpointReferenceType;
import org.ow2.jonas.cxf.transport.reliablejms.api.AsyncHandler;
import org.ow2.jonas.cxf.transport.reliablejms.api.ContextInfo;

public class ReliableJMSTransportConduit extends AbstractConduit implements Conduit, CachedOutputStreamCallback {

    /**
     * Logger.
     */
    private static final Logger LOG = LogUtils.getL7dLogger(ReliableJMSTransportConduit.class);

    /**
     * Atomic number for message count.
     */
    private AtomicLong messageCount;

    /**
     * Correlation ID padding.
     */
    private static final char[] CORRELATTION_ID_PADDING = { '0', '0', '0', '0', '0', '0', '0', '0', '0', '0', '0', '0', '0', '0',
            '0' };

    /**
     * ID.
     */
    private String id = null;

    /**
     * Mode.
     */
    private ReliableJMSTransportModeType mode = null;

    private ReliableJMSTransportConfiguration reliableJMSTransportConfiguration = null;

    public ReliableJMSTransportConduit(ReliableJMSTransportModeType mode, EndpointReferenceType endpointReferenceType,
                                       ReliableJMSTransportConfiguration reliableJMSTransportConfiguration) {
        super(endpointReferenceType);
        this.mode = mode;
        this.reliableJMSTransportConfiguration = reliableJMSTransportConfiguration;
        this.messageCount = new AtomicLong(0);
        this.id = UUID.randomUUID().toString().replaceAll("-", "");
    }

    /**
     * Prepare the given message and this message will be sent when the stream will be closed.
     */
    public void prepare(Message message) throws IOException {

        // Create output stream
        ReliableJMSTransportMessageOutputStream outputStream = new ReliableJMSTransportMessageOutputStream(message.getExchange());

        // Register ourself as a callback
        outputStream.registerCallback(this);

        // Define this in the message
        message.setContent(OutputStream.class, outputStream);
    }

    /**
     * Callback called on the flush method.
     */
    public void onFlush(CachedOutputStream cachedOutputStream) {
        // Do nothing
    }

    /**
     * When the stream is closed, this callback is called. It means that the JMS message may be built and sent
     * @param outputStream the stream containing the data
     */
    public void onClose(CachedOutputStream outputStream) {
        if (!(outputStream instanceof ReliableJMSTransportMessageOutputStream)) {
            throw new IllegalArgumentException("Invalid stream");
        }

        ReliableJMSTransportMessageOutputStream jmsMessageOutputStream = (ReliableJMSTransportMessageOutputStream) outputStream;

        // Send the message if mode is compliant
        switch (mode) {
        case CLIENT_SENDER:
            sendMessage(jmsMessageOutputStream, createCorrelationId(), reliableJMSTransportConfiguration.getTargetDestination(), reliableJMSTransportConfiguration.getReplyToDestination(), reliableJMSTransportConfiguration.getAsyncHandlers());
            break;
        case SERVER_ENDPOINT:
            // Send the message with the same correlationID than the message received and without async handlers (no callback at
            // the endpoint)
            Message message = jmsMessageOutputStream.getExchange().getInMessage();
            sendMessage(jmsMessageOutputStream, (String) message.get("CorrelationID"), (Destination) message.get("JMSReplyTo"), null, null);
            break;

        default:
            // Do nothing for other cases
        }

    }

    public void sendMessage(ReliableJMSTransportMessageOutputStream jmsMessageOutputStream, String correlationID, Destination jmsTarget, Destination jmsReplyTo,
            List<AsyncHandler> asyncHandlers) {

        boolean isTextPayload = ReliableJMSTransportConfiguration.TEXT_MESSAGE_TYPE.equals(reliableJMSTransportConfiguration.getMessageType());
        Object payload = null;
        try {
            payload = retrieveRequestFromStream(jmsMessageOutputStream, isTextPayload);
        } catch (IOException e) {
            throw new IllegalStateException(e);
        }

        Exchange exchange = jmsMessageOutputStream.getExchange();

        // Sending the message that is available in the exchange object
        Message cxfMessage = exchange.getOutMessage();
        
        // If no message, try with the checked fault ?
        boolean isFaultMessage = false;
        if (cxfMessage == null) {
            Message faultMessage = exchange.getOutFaultMessage();
            if (faultMessage != null) {
                FaultMode faultMode = faultMessage.get(org.apache.cxf.message.FaultMode.class);
                if (faultMode == FaultMode.CHECKED_APPLICATION_FAULT) {
                    cxfMessage = faultMessage;
                    isFaultMessage = true;
                }
            }
        }
        
        if (cxfMessage == null) {
            LOG.fine("No output message for the selected operation, so we don't send any message back");
            return;
        }
        
        
        // Needs to send a correct wrapper object to the client
        Message inMessage = exchange.getInMessage();
        Message outMessage = exchange.getOutMessage();

        if (inMessage == null && outMessage != null) {

            // Get invocation context object
            Map<String, Object> context = CastUtils.cast((Map)outMessage.get(Message.INVOCATION_CONTEXT));
            if (context == null) {
                org.apache.cxf.common.i18n.Message msg = new org.apache.cxf.common.i18n.Message("No Invocation context found in the out message '" + outMessage + "'.", getLogger());
                throw new ConfigurationException(msg);
            }
            Map<String, Object> requestContext = CastUtils.cast((Map)context.get(Client.REQUEST_CONTEXT));
            if (requestContext == null) {
                org.apache.cxf.common.i18n.Message msg = new org.apache.cxf.common.i18n.Message("No requestContext found in the invocation context '" + context + "'.", getLogger());
                throw new ConfigurationException(msg);
            }
            
            Method method = (Method) requestContext.get(Method.class.getName());
            if (method == null) {
                org.apache.cxf.common.i18n.Message msg = new org.apache.cxf.common.i18n.Message("No method found in the request context '" + requestContext + "'.", getLogger());
                throw new ConfigurationException(msg);
            }
            
            
            Class<?> returnTypeClass = method.getReturnType();
            
            // Check primitive objects
            // (boolean, byte, char, short, int, long, float, and double)
            List<Object> answerList = new ArrayList<Object>();            
            if (Boolean.TYPE.equals(returnTypeClass)) {
                answerList.add(Boolean.FALSE);
            } else if (Byte.TYPE.equals(returnTypeClass)) {
                answerList.add(Byte.valueOf("0"));
            } else if (Character.TYPE.equals(returnTypeClass)) {
                answerList.add(Character.MIN_VALUE);
            } else if (Short.TYPE.equals(returnTypeClass)) {
                answerList.add(Short.parseShort("0"));
            } else if (Integer.TYPE.equals(returnTypeClass)) {
                answerList.add(Integer.valueOf(0));
            } else if (Long.TYPE.equals(returnTypeClass)) {
                answerList.add(Long.valueOf(0));
            } else if (Float.TYPE.equals(returnTypeClass)) {
                answerList.add(Float.valueOf(0));
            } else if (Double.TYPE.equals(returnTypeClass)) {
                answerList.add(Double.valueOf(0));
            }
            
            // Create a CXF message used to provide the reply to the client (fake answer)
            if (answerList.size() > 0) {
                inMessage = new MessageImpl();
                inMessage.setContent(List.class, answerList);
                exchange.setInMessage(inMessage);
            }
        }

        if (isTextPayload && MessageUtils.isTrue(cxfMessage.getContextualProperty(org.apache.cxf.message.Message.MTOM_ENABLED))
                && cxfMessage.getAttachments() != null && cxfMessage.getAttachments().size() > 0) {
            org.apache.cxf.common.i18n.Message msg = new org.apache.cxf.common.i18n.Message("INVALID_MESSAGE_TYPE", getLogger());
            throw new ConfigurationException(msg);
        }

        // Send correlation ID to async handlers
        if (asyncHandlers != null) {

            ContextInfo contextInfo = new ContextInfo();
            contextInfo.setCorrelationID(correlationID);

            contextInfo.setWsdlService((QName) exchange.get("javax.xml.ws.wsdl.service"));
            contextInfo.setWsdlInterface((QName) exchange.get("javax.xml.ws.wsdl.interface"));
            contextInfo.setWsdlPort((QName) exchange.get("javax.xml.ws.wsdl.port"));
            contextInfo.setWsdlOperation((QName) exchange.get("javax.xml.ws.wsdl.operation"));

            for (AsyncHandler asyncHandler : asyncHandlers) {
                asyncHandler.saveContext(contextInfo);
            }
        }

        // Initialize JMS objects
        Connection connection = null;
        Session session = null;
        MessageProducer messageProducer = null;
        javax.jms.Message jmsMessage = null;

        // create and send the message
        try {

            // Initialize JMS stuff
            connection = reliableJMSTransportConfiguration.getConnectionFactory().createConnection();
            session = connection.createSession(false, Session.AUTO_ACKNOWLEDGE);
            
            // create an empty configuration
            JMSConfiguration jmsConfig = new JMSConfiguration();
            
            jmsMessage = JMSUtils.buildJMSMessageFromCXFMessage(jmsConfig, cxfMessage, payload, reliableJMSTransportConfiguration.getMessageType(),
                    session, jmsReplyTo, correlationID);

            
            // add operation name if this is a fault
            if (isFaultMessage) {
                jmsMessage.setStringProperty(Message.WSDL_OPERATION, ((QName) inMessage.get("javax.xml.ws.wsdl.operation")).toString());
            }
            
            // send it
            messageProducer = session.createProducer(jmsTarget);
            messageProducer.send(jmsMessage);
            //getLogger().fine("Sending the JMS message" + jmsMessage);
        } catch (JMSException e) {
            throw new IllegalStateException("Cannot send the message", e);
        } finally {
            // Close resources
            if (messageProducer != null) {
                try {
                    messageProducer.close();
                } catch (JMSException e) {
                    getLogger().log(Level.SEVERE, "Unable to close message producer", e);
                }
            }
            try {
                if (session != null) {
                    session.close();
                }
            } catch (JMSException e) {
                getLogger().log(Level.SEVERE, "Unable to close session", e);
            }
            try {
                if (connection != null) {
                    connection.close();
                }
            } catch (JMSException e) {
                getLogger().log(Level.SEVERE, "Unable to close connection", e);
            }
        }

    }

    /**
     * Extract the request from the OutputStream
     * @return for textPayloads a String is returned else a byte[]
     * @throws IOException
     */
    protected Object retrieveRequestFromStream(final CachedOutputStream outputStream, boolean isTextPayload) throws IOException {
        Object request = null;
        try {
            if (isTextPayload) {
                StringBuilder builder = new StringBuilder(2048);
                outputStream.writeCacheTo(builder);
                request = builder.toString();
            } else {
                request = outputStream.getBytes();
            }
        } catch (IOException ex) {
            throw new IOException("Error creating request Object from Message content, exception " + ex);
        }
        if (getLogger().isLoggable(Level.FINE)) {
            getLogger().log(Level.FINE, "Payload to be sent out is :[" + request + "]");
        }
        return request;
    }

    protected String createCorrelationId() {
        long i = messageCount.incrementAndGet();
        String index = Long.toHexString(i);
        StringBuilder id = new StringBuilder(this.id);
        id.append(CORRELATTION_ID_PADDING, 0, 16 - index.length());
        id.append(index);
        return id.toString();
    }

    /**
     * @return logger
     */
    protected Logger getLogger() {
        return LOG;
    }
}
