/**
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 2010 France Telecom R&D
 * Contact: jonas-team@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 *
 * --------------------------------------------------------------------------
 * $Id: ReliableJMSTransportFactory.java 21483 2011-06-23 14:47:13Z jppouzol $
 * --------------------------------------------------------------------------
 */

package org.ow2.jonas.cxf.transport.reliablejms;

import java.io.IOException;
import java.lang.reflect.Field;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.logging.Logger;

import javax.annotation.Resource;
import javax.jms.JMSException;
import javax.xml.namespace.QName;

import org.apache.cxf.Bus;
import org.apache.cxf.common.logging.LogUtils;
import org.apache.cxf.configuration.Configurer;
import org.apache.cxf.service.model.BindingInfo;
import org.apache.cxf.service.model.BindingOperationInfo;
import org.apache.cxf.service.model.EndpointInfo;
import org.apache.cxf.service.model.OperationInfo;
import org.apache.cxf.transport.AbstractTransportFactory;
import org.apache.cxf.transport.Conduit;
import org.apache.cxf.transport.ConduitInitiator;
import org.apache.cxf.transport.Destination;
import org.apache.cxf.transport.DestinationFactory;
import org.apache.cxf.ws.addressing.EndpointReferenceType;

/**
 * Transport factory that allows to only send JMS message to a JMS Queue.
 * @author Florent Benoit
 */
public class ReliableJMSTransportFactory extends AbstractTransportFactory implements ConduitInitiator, DestinationFactory {

    /**
     * Logger.
     */
    private static final Logger LOG = LogUtils.getL7dLogger(ReliableJMSTransportFactory.class);

    private static final String PREFIX = "reliablejms://";

    private static final Set<String> URI_PREFIXES = new HashSet<String>();
    static {
        URI_PREFIXES.add(PREFIX);
    }

    private Bus bus;

    private Map<String, ReliableJMSTransportDestination> destinations = null;

    /**
     * outputMessage Field to reset when we want to send messages from the client to the server. By doing this, we're in oneWay
     * everytime and then we don't expect an answer (no wait)
     */
    private static Field outputMessageField = null;

    /**
     * Default constructor.
     */
    public ReliableJMSTransportFactory() {
        this.destinations = new HashMap<String, ReliableJMSTransportDestination>();
        // Initialize once
        if (outputMessageField == null) {
            try {
                outputMessageField = OperationInfo.class.getDeclaredField("outputMessage");
            } catch (SecurityException e) {
                throw new IllegalStateException("Unable to find the given field", e);
            } catch (NoSuchFieldException e) {
                throw new IllegalStateException("Unable to find the given field", e);
            }
            outputMessageField.setAccessible(true);
        }

    }

    /**
     * Sets the CXF Bus
     * @param bus the bus
     */
    @Resource(name = "cxf")
    public void setBus(Bus bus) {
        this.bus = bus;
    }

    /**
     * @return the CXF Bus
     */
    public Bus getBus() {
        return bus;
    }

    /**
     * @return a conduit for the given endpoint info.
     */
    public Conduit getConduit(EndpointInfo endpointInfo) throws IOException {
        return getConduit(getModeType(endpointInfo), disableResponse(endpointInfo), endpointInfo.getTarget());
    }

    /**
     * {@inheritDoc}
     */
    public Conduit getConduit(EndpointInfo endpointInfo, EndpointReferenceType target) throws IOException {
        ReliableJMSTransportModeType mode = getModeType(endpointInfo);
        return getConduit(mode, endpointInfo, target);
    }

    /**
     * Needs to build a destination object based on the given endpoint info. Two cases: endpoint and client receiver Switch is
     * based on the endpoint address
     */
    public Destination getDestination(EndpointInfo endpointInfo) throws IOException {

        ReliableJMSTransportModeType mode = getModeType(endpointInfo);

        switch (mode) {
        case CLIENT_SENDER:
            throw new IllegalStateException("We shouldn't get a destination in mode '" + mode + "' for the endpoint name '"
                    + endpointInfo.getName() + "' with address '" + endpointInfo.getAddress() + "'.");
        case CLIENT_RECEIVER:
        case SERVER_ENDPOINT:

            // build configuration for this endpoint
            ReliableJMSTransportConfiguration reliableJMSTransportConfiguration = getConfiguration(endpointInfo, mode);

            if (reliableJMSTransportConfiguration.getQueue() == null) {
                throw new IllegalStateException("Unable to find configuration for endpoint '" + endpointInfo.getName() + "'.");
            }

            // Build destination
            ReliableJMSTransportDestination reliableJMSTransportDestination = new ReliableJMSTransportDestination(mode, bus, endpointInfo,
                    reliableJMSTransportConfiguration);

            // Check config
            if (reliableJMSTransportConfiguration.getQueue() == null) {
                throw new IllegalStateException("The queue name has not been configured for the destination '"
                        + reliableJMSTransportDestination.getEndpointInfo().getName() + "'.");
            }

            // store this destination in our local map
            String queueName = null;
            try {
                queueName = reliableJMSTransportConfiguration.getQueue().getQueueName();
            } catch (JMSException e) {
                throw new IllegalStateException("Unable to get queue name", e);
            }

            // Check that there is no previous object with the same key
            if (destinations.containsKey(queueName)) {
                throw new IllegalStateException("The queue name '" + queueName + "' is already set for the destination '"
                        + reliableJMSTransportDestination.getEndpointInfo().getName() + "'.");
            }
            LOG.info("Put destination '" + reliableJMSTransportDestination + "' with queue name '" + queueName + "' for instance '"
                    + this + "'.");
            destinations.put(queueName, reliableJMSTransportDestination);

            return reliableJMSTransportDestination;

        default:
            throw new IllegalStateException("Unable to get mode for the endpoint name '" + endpointInfo.getName()
                    + "' with address '" + endpointInfo.getAddress() + "'.");

        }
    }

    protected ReliableJMSTransportModeType getModeType(EndpointInfo endpointInfo) {
        QName endpointName = endpointInfo.getName();
        String address = endpointInfo.getAddress();

        // Find the mode used
        if (address == null || address.length() <= PREFIX.length()) {
            throw new IllegalStateException("Adress wrongly configured for endpoint '" + endpointName + "'. Missing mode");
        }

        ReliableJMSTransportModeType mode = null;
        for (ReliableJMSTransportModeType reliableJMSTransportModeType : ReliableJMSTransportModeType.values()) {
            if (address.contains(reliableJMSTransportModeType.getText())) {
                mode = reliableJMSTransportModeType;
                break;
            }
        }
        if (mode == null) {
            throw new IllegalStateException("Adress wrongly configured for endpoint '" + endpointName + "'. Missing mode");
        }
        return mode;
    }

    /**
     * Remove any output message so that we won't wait when sending requests to the endpoint (oneWay request)
     * @param endpointInfo the given endpoint info
     * @return the modified endpoint info
     */
    protected EndpointInfo disableResponse(EndpointInfo endpointInfo) {
        BindingInfo bindingInfo = endpointInfo.getBinding();
        Collection<BindingOperationInfo> bindingOperationInfos = bindingInfo.getOperations();
        if (bindingOperationInfos != null) {

            // Remove output for each binding operation info
            for (BindingOperationInfo bindingOperationInfo : bindingOperationInfos) {
                OperationInfo operationInfo = bindingOperationInfo.getOperationInfo();
                OperationInfo unwrappedOperationInfo = operationInfo.getUnwrappedOperation();
                try {
                    outputMessageField.set(unwrappedOperationInfo, null);
                } catch (IllegalArgumentException e) {
                    throw new IllegalStateException("Unable to reset output entry", e);
                } catch (IllegalAccessException e) {
                    throw new IllegalStateException("Unable to reset output entry", e);
                }
            }
        }

        return endpointInfo;
    }

    protected Conduit getConduit(ReliableJMSTransportModeType mode, EndpointInfo endpointInfo, EndpointReferenceType target)
            throws IOException {

        ReliableJMSTransportConfiguration reliableJMSTransportConfiguration = getConfiguration(endpointInfo, mode);

        if (reliableJMSTransportConfiguration.getTargetDestination() == null) {
            throw new IllegalStateException("Missing configuration for endpoint '" + endpointInfo.getName() + "'.");
        }

        switch (mode) {
        case CLIENT_SENDER:
            return new ReliableJMSTransportConduit(mode, target, reliableJMSTransportConfiguration);

        default:
            throw new IllegalStateException("We shouldn't expect a conduit for the invalid mode '" + mode
                    + "' for the endpoint name '" + endpointInfo.getName() + "' with address '" + endpointInfo.getAddress()
                    + "'.");

        }

    }

    protected ReliableJMSTransportConfiguration getConfiguration(EndpointInfo endpointInfo, ReliableJMSTransportModeType mode) {
        Configurer configurer = bus.getExtension(Configurer.class);

        // Build the name of the sender configuration
        String name = endpointInfo.getName().toString() + ".reliablejms." + mode.getText() + "-configuration";
        LOG.info("Get configuration for endpoint '" + endpointInfo.getName() + "' with bean configuration name '" + name + "'.");

        ReliableJMSTransportConfiguration reliableJMSTransportConfiguration = new ReliableJMSTransportConfiguration();

        if (configurer != null) {
            configurer.configureBean(name, reliableJMSTransportConfiguration);
        }

        return reliableJMSTransportConfiguration;

    }

    public ReliableJMSTransportDestination getExistingDestination(String queueName) {
        return destinations.get(queueName);
    }

    public Set<String> getUriPrefixes() {
        return URI_PREFIXES;
    }

}